// internal compiler error:
// arm-none-eabi-g++ etc.cpp -c -fomit-frame-pointer -mthumb -fPIC
//
namespace
{

#define CLAMP(ll,x,ul) (((x)<(ll)) ? (ll) : (((x)>(ul)) ? (ul) : (x)))
#define SQUARE(x) ((x)*(x))

#define RED(img,width,x,y)   img[3*(y*width+x)+0]
#define GREEN(img,width,x,y) img[3*(y*width+x)+1]
#define BLUE(img,width,x,y)  img[3*(y*width+x)+2]

#define PUTBITS(dest, data, size, startpos) dest |= ( (data) & ((1<<(size))-1) ) << ((startpos)-(size)+1)
#define PUTBITSHIGH(dest, data, size, startpos) dest |= ( (data) & ((1<<(size))-1) ) << (((startpos)-32)-(size)+1)

static int compressParams[16][4];

int scramble[4] = {3, 2, 0, 1};
int unscramble[4] = {2, 3, 1, 0};

typedef unsigned char uint8;

int compressBlockWithTable2x4(uint8 *img,int width,int height,int startx,int starty,uint8 *avg_color,int table,unsigned int *pixel_indices_MSBp, unsigned int *pixel_indices_LSBp)
{
    uint8 orig[3],approx[3];
    unsigned int pixel_indices_MSB=0, pixel_indices_LSB=0, pixel_indices = 0;
    int sum_error=0;
    int q, i;

    i = 0;
    for (int x=startx; x<startx+2; x++)
    {
        for (int y=starty; y<starty+4; y++)
        {
            int err;
            int best=0;
            int min_error=255*255*3*16;
            orig[0]=RED(img,width,x,y);
            orig[1]=GREEN(img,width,x,y);
            orig[2]=BLUE(img,width,x,y);

            for (q=0;q<4;q++)
            {
                approx[0]=CLAMP(0, avg_color[0]+compressParams[table][q],255);
                approx[1]=CLAMP(0, avg_color[1]+compressParams[table][q],255);
                approx[2]=CLAMP(0, avg_color[2]+compressParams[table][q],255);

                // Here we just use equal weights to R, G and B. Although this will
                // give visually worse results, it will give a better PSNR score.
                err=SQUARE(approx[0]-orig[0]) + SQUARE(approx[1]-orig[1]) + SQUARE(approx[2]-orig[2]);
                if (err<min_error)
                {
                    min_error=err;
                    best=q;
                }
            }

            pixel_indices = scramble[best];

            PUTBITS( pixel_indices_MSB, (pixel_indices >> 1), 1, i);
            PUTBITS( pixel_indices_LSB, (pixel_indices & 1) , 1, i);

            i++;

            // In order to simplify hardware, the table {-12, -4, 4, 12} is indexed {11, 10, 00, 01}
            // so that first bit is sign bit and the other bit is size bit (4 or 12).
            // This means that we have to scramble the bits before storing them.

            sum_error += min_error;
        }
    }

    *pixel_indices_MSBp = pixel_indices_MSB;
    *pixel_indices_LSBp = pixel_indices_LSB;

    return sum_error;
}

}
